#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"

#include "header\font\BGFont.h"
//#include "header\font\BitMapFont.h"

#include "lib.h"
#include "syscall.h"
#include "resource.h"

#include <stdlib.h>

//Mode0
//BG2 ^C
//BG3 \

//---------------------------------------------------------------------------
//2i\L}N 
//http://homepage1.nifty.com/aok2/004/c02.html

#define B8(y)               ( ((0x##y##L & 0x00000001L) >>  0) \
                            | ((0x##y##L & 0x00000010L) >>  3) \
                            | ((0x##y##L & 0x00000100L) >>  6) \
                            | ((0x##y##L & 0x00001000L) >>  9) \
                            | ((0x##y##L & 0x00010000L) >> 12) \
                            | ((0x##y##L & 0x00100000L) >> 15) \
                            | ((0x##y##L & 0x01000000L) >> 18) \
                            | ((0x##y##L & 0x10000000L) >> 21) )
#define B16(h,l)            (B8(h)<<8 | B8(l))
#define B32(hh, hl, lh, ll) (B16(hh, hl)<<16 | B16(lh, ll))

//---------------------------------------------------------------------------
//is
#define STATE_INIT 				1
#define STATE_INPUT				2
#define STATE_RUN  				3
#define STATE_RESET				4

#define MAX_ROM_BYTE_CNT		16
#define MAX_OUT_CNT				4

//---------------------------------------------------------------------------
typedef struct {
	u8   regA;
	u8   regB;
	u8   pc;
	bool cFlg;
} ST_CPU;

typedef struct {
	u8  x;
	u8  y;
	u8  cx;
	u8  cy;
	s16 minX;							//ړ
	s16 maxX;
	s16 minY;
	s16 maxY;
	s16 moveX;
	s16 moveY;
	u16 sprNo;
	u16 sprTile;
	u16 sprSize;
	u16 sprForm;
} ST_CURSOR;

//---------------------------------------------------------------------------
//lib.c
extern ST_TIMER Timer;

//---------------------------------------------------------------------------
ST_BG     BG2;
ST_BG     BG3;
ST_CURSOR Cursor;
u8        State;						//istO

ST_CPU Cpu;
u8     Rom[MAX_ROM_BYTE_CNT];
u8     Out;

//---------------------------------------------------------------------------
void StateInit();
void StateInput();
void StateRun();
void StateReset();

void InitWS();
void InitBG();
void InitSprite();
void InitCursor();
void InitKey();
void IRQUserHandler();

void InitCpu();
void InitRom();
void InitOut();
void InitFrame();

void DrawStatusAll();
void DrawCpuAll();
void DrawRomAll();
void DrawOutAll();
void DrawCursor();
void DrawRow1Bit(u8 x, u8 y, u8 hex);
void DrawRow4Bit(u8 x, u8 y, u8 hex);
void DrawRow8Bit(u8 x, u8 y, u8 hex);
void DrawBool(u8 x, u8 y, bool b);

void ProcKeyInput();
void MoveCursor(u8 key);
bool ChkMoveCursor(u8 key);

void SetRom(u8 adress, u8 bitNum);

void CpuExec();
u8   CpuGetOp(u8 pc);
u8   CpuGetIm(u8 pc);
bool CpuChkCF(u8 reg);

u8   GetCursorxToRomX();
u8   GetCursoryToRomY();

void OmakeDrawPc();
//---------------------------------------------------------------------------
int main()
{
	State = STATE_INIT;

	for(;;)
	{
		WaitForVsync();

		switch(State)
		{
			case STATE_INIT:  StateInit();  break;
			case STATE_INPUT: StateInput(); break;
			case STATE_RUN:   StateRun();   break;
			case STATE_RESET: StateReset(); break;
			default:
				DebugMessage((u8*)"Err: Main State");
				break;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
void StateInit()
{
	InitWS();
	InitBG();
	InitSprite();
	InitCursor();
	InitKey();

	InitCpu();
	InitRom();
	InitOut();
	InitFrame();

	DrawStatusAll();
	DrawCursor();

	SetMode(MODE_0 | BG3_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);
	State = STATE_INPUT;
}
//---------------------------------------------------------------------------
//EFCgRg[
void InitWS()
{
	REG_WSCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT;
}
//---------------------------------------------------------------------------
void InitBG()
{
	//BG3 \

	BG3.chr = 2;
	BG3.map = 1;
	BG3.col = 0;

	REG_BG3CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG3.chr) | BG_MAPBASE(BG3.map) | 0);

	BGSetPal16  (BG3.col, (u16*)&BGFont_palette);
	BGSetChrData(BG3.chr, (u16*)&BGFont_data, BGFont_DATA_SIZE);
}
//---------------------------------------------------------------------------
void InitFrame()
{
	BGDrawCls(BG3.map, ' ');

	BGDrawStr(BG3.map,  0,  0, (u8*)"-- TD4 Emulator --------------");
	BGDrawStr(BG3.map,  0,  1, (u8*)"                              ");
	BGDrawStr(BG3.map,  0,  2, (u8*)" RegA :        00 :           ");
	BGDrawStr(BG3.map,  0,  3, (u8*)" RegB :        01 :           ");
	BGDrawStr(BG3.map,  0,  4, (u8*)" Cflg :        02 :           ");
	BGDrawStr(BG3.map,  0,  5, (u8*)" PC   :        03 :           ");
	BGDrawStr(BG3.map,  0,  6, (u8*)"               04 :           ");
	BGDrawStr(BG3.map,  0,  7, (u8*)" OUT0 :        05 :           ");
	BGDrawStr(BG3.map,  0,  8, (u8*)"    1 :        06 :           ");
	BGDrawStr(BG3.map,  0,  9, (u8*)"    2 :        07 :           ");
	BGDrawStr(BG3.map,  0, 10, (u8*)"    3 :        08 :           ");
	BGDrawStr(BG3.map,  0, 11, (u8*)"               09 :           ");
	BGDrawStr(BG3.map,  0, 12, (u8*)" Clok : 10Hz   10 :           ");
	BGDrawStr(BG3.map,  0, 13, (u8*)"               11 :           ");
	BGDrawStr(BG3.map,  0, 14, (u8*)"               12 :           ");
	BGDrawStr(BG3.map,  0, 15, (u8*)"               13 :           ");
	BGDrawStr(BG3.map,  0, 16, (u8*)"               14 :           ");
	BGDrawStr(BG3.map,  0, 17, (u8*)"               15 :           ");
	BGDrawStr(BG3.map,  0, 18, (u8*)"                              ");
	BGDrawStr(BG3.map,  0, 19, (u8*)"------------------------------");
}
//---------------------------------------------------------------------------
void InitKey()
{
	KeyInit();

	//L[͗p̃^C}ݒ
	REG_INTERUPT = (u32)IRQUserHandler;
	REG_IE       = IRQ_BIT_TIMER0;
	REG_IME      = IRQ_MASTER_ON;

	REG_TM0D     = 65536 - 2621; // (((16*1024*1024) / 64) * 10) / 1000
	REG_TM0CNT   = TM_ENABLE | TM_FREQ_PER_64 | TM_USEIRQ;
}
//---------------------------------------------------------------------------
void IRQUserHandler()
{
	Timer.clockTick += TIMER_INTMS;
	REG_IF = IRQ_BIT_TIMER0;
}
//---------------------------------------------------------------------------
void ProcKeyInput()
{
	if( KeyWaitTimer() == TRUE )
	{
		return;
	}
	if( KeyRepeat() == TRUE )		//L[͂̊i[܂݂܂
	{
		return;
	}

	u8 key = KeyGet();
	if( ChkMoveCursor(key) == TRUE )
	{
		MoveCursor(key);
		DrawCursor();
	}

	if(key & KEY_A)
	{
		u8 bitNum = GetCursorxToRomX();
		u8 adress = GetCursoryToRomY();

		SetRom(adress, bitNum);
		DrawRomAll();
	}

	if(key & KEY_START)
	{
		State = STATE_RUN;
	}
}
//---------------------------------------------------------------------------
bool ChkMoveCursor(u8 key)
{
	ST_CURSOR* p = &Cursor;

	s16 tx1 = p->x;
	s16 ty1 = p->y;

	if(key & KEY_UP   ) ty1 -= p->moveY;
	if(key & KEY_DOWN ) ty1 += p->moveY;
	if(key & KEY_LEFT ) tx1 -= p->moveX;
	if(key & KEY_RIGHT) tx1 += p->moveX;

	if(tx1 < p->minX || ty1 < p->minY) return FALSE;
	if(tx1 > p->maxX || ty1 > p->maxY) return FALSE;

	return TRUE;
}
//---------------------------------------------------------------------------
void MoveCursor(u8 key)
{
	ST_CURSOR* p = &Cursor;

	if(key & KEY_UP   ) p->y -= p->moveY;
	if(key & KEY_DOWN ) p->y += p->moveY;
	if(key & KEY_LEFT ) p->x -= p->moveX;
	if(key & KEY_RIGHT) p->x += p->moveX;
}
//---------------------------------------------------------------------------
void StateInput()
{
	ProcKeyInput();
}
//---------------------------------------------------------------------------
void StateRun()
{
	CpuExec();
}
//---------------------------------------------------------------------------
void StateReset()
{
	
}
//---------------------------------------------------------------------------
void InitCpu()
{
	ST_CPU* p = &Cpu;

	p->regA   = 0x00;
	p->regB   = 0x00;
	p->pc     = 0x00;
	p->cFlg  = FALSE;
}
//---------------------------------------------------------------------------
void InitRom()
{
	//[^C}[iXs[h10Hẑ10{j
	//             Jn`1
	//             1`2
	//   2`3
	//             3

	Rom[ 0] = B8(10110111);
	Rom[ 1] = B8(00000001);
	Rom[ 2] = B8(11100001);
	Rom[ 3] = B8(00000001);
	Rom[ 4] = B8(11100011);
	Rom[ 5] = B8(10110110);
	Rom[ 6] = B8(00000001);
	Rom[ 7] = B8(11100110);
	Rom[ 8] = B8(00000001);
	Rom[ 9] = B8(11101000);
	Rom[10] = B8(10110000);
	Rom[11] = B8(10110100);
	Rom[12] = B8(00000001);
	Rom[13] = B8(11101010);
	Rom[14] = B8(10111000);
	Rom[15] = B8(11111111);
}
//---------------------------------------------------------------------------
void InitOut()
{
	Out = 0x00;
}
//---------------------------------------------------------------------------
void DrawStatusAll()
{
	DrawCpuAll();
	DrawRomAll();
	DrawOutAll();
}
//---------------------------------------------------------------------------
void DrawCpuAll()
{
	ST_CPU* p = &Cpu;

	DrawRow4Bit(8, 2, p->regA );
	DrawRow4Bit(8, 3, p->regB );
	DrawRow1Bit(8, 4, p->cFlg);
	DrawRow4Bit(8, 5, p->pc   );
}
//---------------------------------------------------------------------------
void DrawRomAll()
{
	s16 i;
	for(i=0; i<MAX_ROM_BYTE_CNT; i++)
	{
		DrawRow8Bit(20, 2+i, Rom[i]);
	}
}
//---------------------------------------------------------------------------
void DrawOutAll()
{
	DrawBool(8,  7, (Out & 0x01) >> 0);
	DrawBool(8,  8, (Out & 0x02) >> 1);
	DrawBool(8,  9, (Out & 0x04) >> 2);
	DrawBool(8, 10, (Out & 0x08) >> 3);
}
//---------------------------------------------------------------------------
void DrawRow8Bit(u8 x, u8 y, u8 hex)
{
	u8  mask = 0x80;

	s16 i;
	for(i=0; i<8; i++)
	{
		if(hex & mask) BGDrawChar(BG3.map, x+i, y, BG3.col, '1');
		else           BGDrawChar(BG3.map, x+i, y, BG3.col, '0');

		mask >>= 1;
	}
}
//---------------------------------------------------------------------------
void DrawRow4Bit(u8 x, u8 y, u8 hex)
{
	u8  mask = 0x08;

	s16 i;
	for(i=0; i<4; i++)
	{
		if(hex & mask) BGDrawChar(BG3.map, x+i, y, BG3.col, '1');
		else           BGDrawChar(BG3.map, x+i, y, BG3.col, '0');

		mask >>= 1;
	}
}
//---------------------------------------------------------------------------
void DrawRow1Bit(u8 x, u8 y, u8 hex)
{
	u8 mask = 0x01;

	if(hex & mask) BGDrawChar(BG3.map, x, y, BG3.col, '1');
	else           BGDrawChar(BG3.map, x, y, BG3.col, '0');
}
//---------------------------------------------------------------------------
void DrawBool(u8 x, u8 y, bool b)
{
	if(b == TRUE) BGDrawStr(BG3.map,  x,  y, (u8*)"ON ");
	else          BGDrawStr(BG3.map,  x,  y, (u8*)"OFF");
}
//---------------------------------------------------------------------------
void InitSprite()
{
	SpriteInit();
	SpriteSetPal16((u16*)&cursor_pal_bin);
	SpriteSetData ((u16*)&cursor_img_bin, CURSOR_IMG_SIZE);
}
//---------------------------------------------------------------------------
void InitCursor()
{
	ST_CURSOR* cp = &Cursor;

	cp->x       = 20;
	cp->y       =  2;
	cp->cx      =  8;
	cp->cy      =  8;
	cp->minX    = 20;			//ړ
	cp->maxX    = 27;
	cp->minY    =  2;
	cp->maxY    = 17;
	cp->moveX   =  1;
	cp->moveY   =  1;
	cp->sprNo   =  0;
	cp->sprTile =  0;
	cp->sprSize = SP_SIZE_8;
	cp->sprForm = SP_SQUARE;

	SpriteSetSize(Cursor.sprNo, Cursor.sprSize, Cursor.sprForm, SP_COLOR_16);
	SpriteSetChr (Cursor.sprNo, Cursor.sprTile);
}
//---------------------------------------------------------------------------
void DrawCursor()
{
	s16 x  = Cursor.x * Cursor.cx;
	s16 y  = Cursor.y * Cursor.cy;
	u16 no = Cursor.sprNo;

	SpriteMove(no, x, y);
}
//---------------------------------------------------------------------------
u8 GetCursorxToRomX()
{
	return Cursor.x - 20;
}
//---------------------------------------------------------------------------
u8 GetCursoryToRomY()
{
	return Cursor.y - 2;
}
//---------------------------------------------------------------------------
void SetRom(u8 adress, u8 bitNum)
{
	u8 mask = 0x80;
	mask >>= bitNum;

	if(mask & Rom[adress]) Rom[adress] &= ~mask;
	else                   Rom[adress] |=  mask;
}
//---------------------------------------------------------------------------
void CpuExec()
{
	InitCpu();		//CpuReset

	u8 op, Im;
	ST_CPU* p = &Cpu;

	for(;;)
	{
		DrawCpuAll();
		DrawOutAll();
		OmakeDrawPc();

		op    = CpuGetOp(p->pc);
		Im    = CpuGetIm(p->pc);
		p->pc = (p->pc + 1) & 0x0f;

		switch(op)
		{

		//MOV A,Im
		case 0x03:
			p->regA  = Im;
			p->cFlg  = FALSE;
			break;

		//MOV B,Im
		case 0x07:
			p->regB  = Im;
			p->cFlg  = FALSE;
			break;

		//MOV A,B
		case 0x01:
			p->regA  = p->regB;
			p->cFlg  = FALSE;
			break;

		//MOV B,A
		case 0x04:
			p->regB  = p->regA;
			p->cFlg  = FALSE;
			break;

		//ADD A,Im
		case 0x00:
			p->regA += Im;
			p->cFlg  = CpuChkCF(p->regA);
			p->regA &= 0x0F;
			break;

		//ADD B,Im
		case 0x05:
			p->regB += Im;
			p->cFlg  = CpuChkCF(p->regB);
			p->regB &= 0x0F;
			break;

		case 0x02:	//IN A(ĂȂł)
			break;

		case 0x06:	//IN B(ĂȂł)
			break;

		case 0x0b:	//OUT Im
			Out      = Im;
			p->cFlg  = FALSE;
			break;

		case 0x09:	//OUT B
			Out      = p->regB;
			p->cFlg  = FALSE;
			break;

		//JMP Im
		case 0x0f:
			p->pc    = Im;
			p->cFlg  = FALSE;
			break;

		//JNC Im
		case 0x0e:
			if(p->cFlg == FALSE)
			{
				p->pc = Im;
			}
			p->cFlg  = FALSE;
			break;

		default: 
			break;
		}

		WaitMs(100);	//10Hz
	}
}
//---------------------------------------------------------------------------
bool CpuChkCF(u8 reg)
{
	return reg & 0xf0 ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
u8 CpuGetOp(u8 pc)
{
	return Rom[pc] >> 4;
}
//---------------------------------------------------------------------------
u8 CpuGetIm(u8 pc)
{
	return Rom[pc] & 0x0f;
}
//---------------------------------------------------------------------------
void OmakeDrawPc()
{
	Cursor.x = 14;
	Cursor.y = Cpu.pc + 2;

	DrawCursor();
}
